#!/usr/bin/env python

"""
Copyright (c) 2006-2024 sqlmap developers (https://sqlmap.org/)
See the file 'LICENSE' for copying permission
"""

# 导入必要的模块
from __future__ import division  # 确保除法运算返回浮点数

import time  # 用于计时功能

# 导入自定义模块
from lib.core.common import dataToStdout  # 用于向标准输出写入数据
from lib.core.convert import getUnicode  # 用于字符串Unicode转换
from lib.core.data import conf  # 配置相关
from lib.core.data import kb  # 知识库相关

class ProgressBar(object):
    """
    这个类定义了更新和绘制进度条的方法
    """

    def __init__(self, minValue=0, maxValue=10, totalWidth=None):
        """
        初始化进度条
        参数:
            minValue: 最小值(默认0)
            maxValue: 最大值(默认10)
            totalWidth: 进度条总宽度(如果为None则使用配置中的宽度)
        """
        self._progBar = "[]"  # 进度条的基本形状
        self._min = int(minValue)  # 最小值
        self._max = int(maxValue)  # 最大值
        self._span = max(self._max - self._min, 0.001)  # 计算范围(避免除以0)
        self._width = totalWidth if totalWidth else conf.progressWidth  # 设置进度条宽度
        self._amount = 0  # 当前进度值
        self._start = None  # 开始时间
        self.update()  # 初始化时更新一次

    def _convertSeconds(self, value):
        """
        将秒数转换为分:秒格式
        例如: 90秒 -> "01:30"
        """
        seconds = value
        minutes = seconds // 60  # 计算分钟数
        seconds = seconds - (minutes * 60)  # 计算剩余秒数

        return "%.2d:%.2d" % (minutes, seconds)  # 返回格式化的时间字符串

    def update(self, newAmount=0):
        """
        更新进度条的状态
        参数:
            newAmount: 新的进度值
        """
        # 确保进度值在合理范围内
        if newAmount < self._min:
            newAmount = self._min
        elif newAmount > self._max:
            newAmount = self._max

        self._amount = newAmount

        # 计算完成百分比
        diffFromMin = float(self._amount - self._min)
        percentDone = (diffFromMin / float(self._span)) * 100.0
        percentDone = round(percentDone)
        percentDone = min(100, int(percentDone))

        # 计算需要显示多少个等号作为进度标记
        allFull = self._width - len("100%% [] %s/%s  (ETA 00:00)" % (self._max, self._max))
        numHashes = (percentDone / 100.0) * allFull
        numHashes = int(round(numHashes))

        # 构建进度条字符串
        if numHashes == 0:  # 0%进度
            self._progBar = "[>%s]" % (" " * (allFull - 1))
        elif numHashes == allFull:  # 100%进度
            self._progBar = "[%s]" % ("=" * allFull)
        else:  # 部分进度
            self._progBar = "[%s>%s]" % ("=" * (numHashes - 1), " " * (allFull - numHashes))

        # 在进度条开头添加百分比
        percentString = getUnicode(percentDone) + "%"
        self._progBar = "%s %s" % (percentString, self._progBar)

    def progress(self, newAmount):
        """
        保存时间增量并显示更新后的进度条及预计完成时间
        参数:
            newAmount: 新的进度值
        """
        # 计算预计完成时间(ETA)
        if self._start is None or newAmount > self._max:
            self._start = time.time()
            eta = None
        else:
            delta = time.time() - self._start  # 计算已经过时间
            eta = (self._max - self._min) * (1.0 * delta / newAmount) - delta  # 估算剩余时间

        self.update(newAmount)  # 更新进度
        self.draw(eta)  # 绘制进度条

    def draw(self, eta=None):
        """
        如果进度发生变化则绘制进度条
        参数:
            eta: 预计完成时间(秒)
        """
        # 在同一行更新进度条显示
        dataToStdout("\r%s %d/%d%s" % (
            self._progBar,  # 进度条
            self._amount,   # 当前进度
            self._max,      # 最大值
            ("  (ETA %s)" % (self._convertSeconds(int(eta)) if eta is not None else "??:??"))  # 显示预计完成时间
        ))
        
        # 当进度达到最大值时清除进度条
        if self._amount >= self._max:
            dataToStdout("\r%s\r" % (" " * self._width))
            kb.prependFlag = False

    def __str__(self):
        """
        返回进度条的字符串表示
        """
        return getUnicode(self._progBar)
