#!/usr/bin/env python

"""
Copyright (c) 2006-2024 sqlmap developers (https://sqlmap.org/)
See the file 'LICENSE' for copying permission
"""

import re# 导入正则表达式模块

from lib.core.data import kb  # 导入知识库（包含SQL关键字等信息）
from lib.core.enums import PRIORITY # 导入优先级枚举

# 设置优先级为普通
__priority__ = PRIORITY.NORMAL

def dependencies():
    pass

def tamper(payload, **kwargs):
    """
    这个函数用于篡改（tamper）输入的payload，通过替换空格字符和等号来绕过Blue Coat SGOS的WAF。

    功能：
        * 将SQL语句后的空格替换为有效的随机空白字符。
        * 将等号（=）替换为LIKE操作符。

    要求：
        * Blue Coat SGOS，且WAF已激活，参考文档：https://kb.bluecoat.com/index?page=content&id=FAQ2147

    测试情况：
        * MySQL 5.1, SGOS

    注意：
        * 这个篡改方法对于绕过Blue Coat推荐的WAF规则配置很有用。

    示例：
        >>> tamper('SELECT id FROM users WHERE id = 1')
        'SELECT%09id FROM%09users WHERE%09id LIKE 1'
    """

    def process(match):
        """
        辅助函数，用于处理正则匹配的结果。
        将匹配到的SQL关键字替换为带有%09（制表符）的版本。
        """
        word = match.group('word')
        if word.upper() in kb.keywords:
            return match.group().replace(word, "%s%%09" % word)
        else:
            return match.group()

    retVal = payload # 初始化返回值

    if payload:
        # 替换SQL关键字后跟非单词字符或字符串末尾的空格
        retVal = re.sub(r"\b(?P<word>[A-Z_]+)(?=[^\w(]|\Z)", process, retVal)
         # 将等号替换为LIKE操作符
        retVal = re.sub(r"\s*=\s*", " LIKE ", retVal)
        # 移除多余的%09空格
        retVal = retVal.replace("%09 ", "%09")

    return retVal
