#!/usr/bin/env python

"""
Copyright (c) 2006-2024 sqlmap developers (https://sqlmap.org/)
See the file 'LICENSE' for copying permission
"""

from lib.core.compat import xrange  # 导入兼容库中的xrange函数，用于兼容Python 2和3的range函数
from lib.core.enums import PRIORITY  # 从核心库导入优先级枚举
# 设置优先级为最高
__priority__ = PRIORITY.HIGHEST

def dependencies():
    pass

def tamper(payload, **kwargs):
    """
    这个函数用于篡改（tamper）输入的payload，将'IFNULL(A, B)'语句替换为其等效的'IF(ISNULL(A), B, A)'形式。

    参数：
        payload：要篡改的原始payload。
        **kwargs：其他可选参数（在本函数中未使用）。

    要求：
        * 适用于MySQL、SQLite（可能）和SAP MaxDB（可能）数据库。

    测试情况：
        * MySQL 5.0 和 5.5

    注意：
        * 这个篡改方法对于绕过那些过滤IFNULL()函数的非常弱的定制Web应用防火墙很有用。

    示例：
        >>> tamper('IFNULL(1, 2)')
        'IF(ISNULL(1),2,1)'
    """

    if payload and payload.find("IFNULL") > -1:  # 如果payload不为空且包含'IFNULL'
        while payload.find("IFNULL(") > -1:  # 遍历所有'IFNULL'语句
            index = payload.find("IFNULL(") # 找到'IFNULL'的位置
            depth = 1 # 初始化括号深度
            comma, end = None, None  # 初始化逗号位置和结束位置

            # 遍历payload以找到'IFNULL'语句的结束位置
            for i in xrange(index + len("IFNULL("), len(payload)):
                if depth == 1 and payload[i] == ',':
                    comma = i  # 记录逗号位置

                elif depth == 1 and payload[i] == ')':
                    end = i  # 记录结束位置
                    break

                elif payload[i] == '(':
                    depth += 1  # 增加括号深度

                elif payload[i] == ')':
                    depth -= 1  # 减少括号深度

            # 如果找到逗号和结束位置，则进行替换
            if comma and end:
                _ = payload[index + len("IFNULL("):comma]  # 提取参数A
                __ = payload[comma + 1:end].lstrip() # 提取参数B
                newVal = "IF(ISNULL(%s),%s,%s)" % (_, __, _)  # 构造新的IF语句
                payload = payload[:index] + newVal + payload[end + 1:]  # 替换原IFNULL语句
            else:
                break

    return payload
